@@ -1,6 +1,6 @@
 use v6;
 
-unit module TAP:ver<0.3.7>;
+unit module TAP:ver<0.3.9>;
 
 role Entry {
     has Str:D $.raw is required handles <Str>;
@@ -107,7 +107,7 @@ class Result {
     has Duration $.time;
 
     method has-problems($ignore-exit = False) {
-        so @!failed || @!errors || (!$ignore-exit && self.exit-failed);
+        so @!failed || @!errors || (!$ignore-exit && $!exit-status.wait);
     }
 }
 
@@ -166,12 +166,13 @@ class Aggregator {
     }
 }
 
-my grammar Grammar {
-    regex TOP {
-        ^ [ <plan> | <test> | <bailout> | <version> | <pragma> | <comment> || <unknown> ] $
-    }
+grammar Grammar {
+    token TOP { [ <entry> \n ]+ }
     token sp { <[\s] - [\n]> }
     token num { <[0..9]>+ }
+    token entry {
+        ^^ [ <plan> | <test> | <bailout> | <version> | <comment> | <pragma> | <yaml> | <sub-test> || <unknown> ] $$
+    }
     token plan {
         '1..' <count=.num> <.sp>* [
             '#' <.sp>* $<directive>=[:i 'SKIP'] \S*
@@ -224,73 +225,69 @@ my grammar Grammar {
     token unknown {
         \N*
     }
-    class Actions {
-        method TOP($/) {
-            make $/.values[0].made;
-        }
-        method plan($/) {
-            my %args = :raw(~$/), :tests(+$<count>);
-            if $<directive> {
-                %args<skip-all explanation> = True, ~$<explanation>;
-            }
-            make TAP::Plan.new(|%args);
-        }
-        method description($m) {
-            $m.make: ~$m.subst(/\\('#'|'\\')/, { $_[0] }, :g)
-        }
-        method !make_test($/) {
-            my %args = (:ok($<nok> eq ''));
-            %args<number> = $<num> ?? +$<num> !! Int;
-            %args<description> = $<description>.made if $<description>;
-            %args<directive> = $<directive> ?? TAP::Directive::{~$<directive>.substr(0,4).tclc} !! TAP::No-Directive;
-            %args<explanation> = ~$<explanation> if $<explanation>;
-            %args;
-        }
-        method test($/) {
-            make TAP::Test.new(:raw(~$/), |self!make_test($/));
-        }
-        method bailout($/) {
-            make TAP::Bailout.new(:raw(~$/), :explanation($<explanation> ?? ~$<explanation> !! Str));
-        }
-        method version($/) {
-            make TAP::Version.new(:raw(~$/), :version(+$<version>));
-        }
-        method pragma-identifier($/) {
-            make $<name> => ?( $<sign> eq '+' );
-        }
-        method pragma($/) {
-            my Bool:D %identifiers = @<pragma-identifier>.map(*.ast);
-            make TAP::Pragma.new(:raw(~$/), :%identifiers);
-        }
-        method comment($/) {
-            make TAP::Comment.new(:raw(~$/), :comment(~$<comment>));
-        }
-        method yaml($/) {
-            my $serialized = $<yaml-line>.join('');
-            my $deserialized = try (require YAMLish) ?? YAMLish::load-yaml("---\n$serialized...") !! Any;
-            make TAP::YAML.new(:raw(~$/), :$serialized, :$deserialized);
-        }
-        method sub-entry($/) {
-            make $/.values[0].made;
-        }
-        method sub-test($/) {
-            make TAP::Sub-Test.new(:raw(~$/), :entries(@<sub-entry>».made), |self!make_test($<test>));
-        }
-        method unknown($/) {
-            make TAP::Unknown.new(:raw(~$/));
-        }
+}
+
+class Actions {
+    method TOP($/) {
+        make @<entry>».made;
     }
-    method parse(|c) {
-        my $*tap-indent = 0;
-        nextwith(:actions(Actions), |c);
+    method entry($/) {
+        make $/.values[0].made;
     }
-    method subparse(|c) {
-        my $*tap-indent = 0;
-        nextwith(:actions(Actions), |c);
+    method plan($/) {
+        my %args = :raw(~$/), :tests(+$<count>);
+        if $<directive> {
+            %args<skip-all explanation> = True, ~$<explanation>;
+        }
+        make TAP::Plan.new(|%args);
+    }
+    method description($m) {
+        $m.make: ~$m.subst(/\\('#'|'\\')/, { $_[0] }, :g)
+    }
+    method !make_test($/) {
+        my %args = (:ok($<nok> eq ''));
+        %args<number> = $<num> ?? +$<num> !! Int;
+        %args<description> = $<description>.made if $<description>;
+        %args<directive> = $<directive> ?? TAP::Directive::{~$<directive>.substr(0,4).tclc} !! TAP::No-Directive;
+        %args<explanation> = ~$<explanation> if $<explanation>;
+        %args;
+    }
+    method test($/) {
+        make TAP::Test.new(:raw(~$/), |self!make_test($/));
+    }
+    method bailout($/) {
+        make TAP::Bailout.new(:raw(~$/), :explanation($<explanation> ?? ~$<explanation> !! Str));
+    }
+    method version($/) {
+        make TAP::Version.new(:raw(~$/), :version(+$<version>));
+    }
+    method pragma-identifier($/) {
+        make $<name> => ?( $<sign> eq '+' );
+    }
+    method pragma($/) {
+        my Bool:D %identifiers = @<pragma-identifier>.map(*.ast);
+        make TAP::Pragma.new(:raw(~$/), :%identifiers);
+    }
+    method comment($/) {
+        make TAP::Comment.new(:raw(~$/), :comment(~$<comment>));
+    }
+    method yaml($/) {
+        my $serialized = $<yaml-line>.join('');
+        my $deserialized = try (require YAMLish) ?? YAMLish::load-yaml("---\n$serialized...") !! Any;
+        make TAP::YAML.new(:raw(~$/), :$serialized, :$deserialized);
+    }
+    method sub-entry($/) {
+        make $/.values[0].made;
+    }
+    method sub-test($/) {
+        make TAP::Sub-Test.new(:raw(~$/), :entries(@<sub-entry>».made), |self!make_test($<test>));
+    }
+    method unknown($/) {
+        make TAP::Unknown.new(:raw(~$/));
     }
 }
 
-my sub parser(Supply $input --> Supply) {
+my sub parse-stream(Supply $input --> Supply) {
     supply {
         enum Mode <Normal SubTest Yaml >;
         my Mode $mode = Normal;
@@ -324,15 +321,15 @@ my sub parser(Supply $input --> Supply) {
                 } elsif $line ~~ &indented {
                     set-state(SubTest, $line);
                 } else {
-                    emit-reset $grammar.parse($line);
+                    emit-reset $grammar.parse($line, :rule('entry'), :actions(Actions));
                 }
             }
             elsif $mode == SubTest {
                 if $line ~~ &indented {
                     @buffer.push: $line;
-                } elsif $grammar.parse($line, :rule('test')) -> $test {
+                } elsif $grammar.parse($line, :rule('test'), :actions(Actions)) -> $test {
                     my $raw = (|@buffer, $line).join("\n");
-                    if $grammar.parse($raw, :rule('sub-test')) -> $subtest {
+                    if $grammar.parse($raw, :rule('sub-test'), :actions(Actions)) -> $subtest {
                         emit-reset $subtest;
                     } else {
                         emit-unknown;
@@ -347,7 +344,7 @@ my sub parser(Supply $input --> Supply) {
                     @buffer.push: $line;
                     if $<content> eq '...' {
                         my $raw = @buffer.join("\n");
-                        if $grammar.parse($raw, :rule('yaml')) -> $yaml {
+                        if $grammar.parse($raw, :rule('yaml'), :actions(Actions)) -> $yaml {
                             emit-reset $yaml;
                         } else {
                             emit-unknown;
@@ -362,7 +359,14 @@ my sub parser(Supply $input --> Supply) {
     }
 }
 
-role Output {
+role Output does Entry::Handler {
+    method handle-entry(Entry $entry) {
+        self.say(~$entry)
+    }
+    method end-entries() {
+        self.flush;
+    }
+
     method print(Str $value) {
         ...
     }
@@ -687,6 +691,10 @@ my class State does TAP::Entry::Handler {
     has Int $!version;
     has Bool $.loose;
 
+    submethod TWEAK(Supply :$events) {
+        self.listen($events);
+    }
+
     proto method handle-entry(TAP::Entry $entry) {
         if $!seen-plan == After && $entry !~~ TAP::Comment {
             self!add-error("Got line $entry after late plan");
@@ -778,29 +786,22 @@ my class State does TAP::Entry::Handler {
     }
 }
 
-my class Run {
+class Parser does Awaitable {
     subset Killable of Any where { .can('kill') };
+
+    has Str $.name;
     has Supply:D $.events is required;
-    my sub get-finished(Supply $supply) {
-        my $result = Promise.new;
-        my &done = -> { $result.keep(Status.new(0, 0)) };
-        my &quit = -> $reason { $result.keep(Status.new(255, 0)) };
-        $supply.tap(-> $ {}, :&done, :&quit);
-        $result;
-    }
-    has Promise:D $.process = get-finished($!events);
+    has Promise:D $.process = $!events.Promise.then({ Status.new($^promise.status ~~ Kept ?? 0 !! 255, 0) });
     has Killable $!killer;
     has Promise $!timer;
+    has State $!state is built;
+    has Promise $.promise is built(False) handles <result get-await-handle> = Promise.allof($!state.done, $!process).then({ $!state.finalize($!name, self.exit-status, self.time) });
 
     method kill() {
         $!killer.kill if $!process;
     }
     method exit-status() {
-        with try $!process.result -> $result {
-            Status.new($result.exit, $result.signal);
-        } else {
-            Status.new(255, 0);
-        }
+        try { $!process.result } // Status.new(255, 0);
     }
     method time() {
         $!timer.defined ?? $!timer.result !! Duration;
@@ -808,36 +809,21 @@ my class Run {
 }
 
 role Source {
-    has Str $.name;
+    has Str $.name = '';
+    method parse(Promise :$bailout, Bool :$loose) { ... }
 }
 class Source::Proc does Source {
     has Str $.path is required;
     has @.args;
     has $.err is required;
     has $.cwd is required;
-}
-class Source::File does Source {
-    has IO(Str) $.filename handles<slurp>;
-}
-class Source::String does Source {
-    has Str $.content;
-}
-class Source::Supply does Source {
-    has Supply $.supply;
-}
 
-class Async {
-    has Str $.name;
-    has Run $!run handles <kill events> is built;
-    has State $!state is built;
-    has Promise $.waiter is built(False) = Promise.allof($!state.done, $!run.process);
-
-    multi get_runner(Source::Proc $proc) {
-        my $async = Proc::Async.new($proc.path, $proc.args);
-        my $events = parser($async.stdout);
+    method parse(Promise :$bailout, Bool :$loose) {
+        my $async = Proc::Async.new($!path, @!args);
+        my $events = parse-stream($async.stdout);
         state $devnull;
         END { $devnull.close with $devnull }
-        given $proc.err {
+        given $!err {
             my $err = $_;
             when 'stderr' { #default is correct
             }
@@ -860,76 +846,125 @@ class Async {
                 die "Unknown error handler";
             }
         }
-        my $process = $async.start(:cwd($proc.cwd));
+        my $state = State.new(:$bailout, :$loose, :$events);
+        my $start = $async.start(:$!cwd);
+        my $process = $start.then({ my $result = $start.result; Status.new($result.exitcode, $result.signal) });
         my $start-time = now;
         my $timer = $process.then({ now - $start-time });
-        Run.new(:$process, :killer($async), :$timer, :$events);
+        Parser.new(:$!name, :$state, :$process, :killer($async), :$timer, :$events);
     }
-    multi get_runner(Source::Supply $supply) {
+}
+class Source::File does Source {
+    has IO(Str) $.filename handles<slurp>;
+
+    method parse(Promise :$bailout, Bool :$loose) {
+        my $events = parse-stream(supply { emit self.slurp(:close) });
+        my $state = State.new(:$bailout, :$loose, :$events);
+        Parser.new(:$!name, :$state, :$events);
+    }
+}
+class Source::String does Source {
+    has Str $.content;
+
+    method parse(Promise :$bailout, Bool :$loose) {
+        my $events = parse-stream(supply { emit $!content });
+        my $state = State.new(:$bailout, :$loose, :$events);
+        Parser.new(:$!name, :$state, :$events);
+    }
+}
+class Source::Supply does Source {
+    has Supply $.supply;
+
+    method parse(Promise :$bailout, Bool :$loose) {
         my $start-time = now;
-        my $events = parser($supply.supply);
-        Run.new(:$events);
-    }
-    multi get_runner(Source::File $file) {
-        my $events = parser(supply { emit $file.slurp(:close) });
-        Run.new(:$events);
-    }
-    multi get_runner(Source::String $string) {
-        my $events = parser(supply { emit $string.content });
-        Run.new(:$events);
+        my $events = parse-stream($!supply);
+        my $state = State.new(:$bailout, :$loose, :$events);
+        my $timer = $events.Promise.then( -> $ { now - $start-time });
+        Parser.new(:$!name, :$state, :$events, :$timer);
     }
+}
 
-    method new(Source :$source, Promise :$bailout, Bool :$loose) {
-        my $state = State.new(:$bailout, :$loose);
-        my $run = get_runner($source);
-        $state.listen($run.events);
-        Async.bless(:name($source.name), :$state, :$run);
-    }
+subset SourceHandler::Priority of Numeric where 0..1;
 
-    has TAP::Result $!result;
-    method result {
-        await $!waiter;
-        $!result //= $!state.finalize($!name, $!run.exit-status, $!run.time);
+role SourceHandler {
+    method can-handle {...};
+    method make-source {...};
+
+    multi method make-parser(Str:D $name, Any:D :$err = 'stderr', IO:D(Str) :$cwd = $*CWD, Promise :$bailout, Bool :$loose, *%args) {
+        my $source = self.make-source($name, :$err, :$cwd, |%args);
+        $source.parse(:$bailout, :$loose);
+    }
+    multi method make-parser(IO:D $path, IO:D(Str) :$cwd = $path.CWD, *%args) {
+        self.make-parser($path.relative($cwd), :$cwd, |%args);
+    }
+}
+
+my sub normalize-path($path, IO::Path $cwd) {
+    $path ~~ IO ?? $path.relative($cwd) !! ~$path
+}
+
+class SourceHandler::Raku does SourceHandler {
+    has Str:D $.path = $*EXECUTABLE.absolute;
+    has @.incdirs;
+    method make-source(Str:D $name, Any:D :$err, IO::Path:D :$cwd, :@include-dirs = (), *%) {
+        my @dirs = flat @include-dirs, @!incdirs;
+        my @args = |@dirs.map({ "-I" ~ normalize-path($^dir, $cwd) }), $name;
+        TAP::Source::Proc.new(:$name, :$!path, :@args, :$err, :$cwd);
+    }
+    method can-handle(Str $name) {
+        0.5;
+    }
+}
+
+class SourceHandler::Exec does SourceHandler {
+    has @.args;
+    has SourceHandler::Priority $.priority = 1;
+    method new (*@args) {
+        self.bless(:@args);
+    }
+    method can-handle(Str $name) {
+        $!priority;
+    }
+    method make-source(Str:D $name, Any:D :$err, IO::Path:D :$cwd, *%) {
+        my $executable = ~$cwd.add($name);
+        my ($path, *@args) = @!args ?? (|@!args, $executable) !! $executable;
+        TAP::Source::Proc.new(:$name, :$path, :@args, :$err, :$cwd);
+    }
+}
+
+class SourceHandler::File does SourceHandler {
+    method can-handle(Str $name) {
+        $name ~~ /\.tap$/ ?? 1 !! 0;
+    }
+    method make-source(Str:D $name, Any:D :$err, IO::Path:D :$cwd, *%) {
+        my $filename = $cwd.add($name);
+        TAP::Source::File.new(:$name, :$filename);
+    }
+}
+
+class SourceHandlers {
+    has TAP::SourceHandler @.handlers = ( SourceHandler::Raku.new, TAP::SourceHandler::File.new );
+    multi method make-source(Str:D $path, *%args) {
+        @!handlers.max(*.can-handle($path)).make-source($path, |%args)
+    }
+    multi method make-source(IO:D $path, IO:D(Str) :$cwd = $path.CWD, *%args) {
+        self.make-source($path.relative($cwd), :$cwd, |%args);
+    }
+    multi method make-parser(Any:D $path, Promise :$bailout, Bool :$loose, *%args) {
+        my $source = self.make-source($path, |%args);
+        $source.parse(:$bailout, :$loose);
+    }
+    method COERCE($handlers) {
+        self.new(:handlers($handlers.list));
     }
 }
 
 class Harness {
-    role SourceHandler {
-        method can-handle {...};
-        method make-source {...};
-    }
-    role SourceHandler::Proc does SourceHandler {
-    }
-    my sub normalize-path($path, IO::Path $cwd) {
-        $path ~~ IO ?? $path.IO.relative($cwd) !! ~$path
-    }
-    class SourceHandler::Raku does SourceHandler::Proc {
-        has Str:D $.path = $*EXECUTABLE.absolute;
-        has @.incdirs;
-        method make-source(Str:D $name, Any:D :$err, IO::Path:D :$cwd, :@include-dirs = (), *%) {
-            my @dirs = map { normalize-path($_, $cwd) }, |@include-dirs, |@!incdirs;
-            my @args = |@dirs.map("-I" ~ *), $name;
-            TAP::Source::Proc.new(:$name, :$!path, :@args, :$err, :$cwd);
-        }
-        method can-handle(Str $name) {
-            0.5;
-        }
-    }
-    class SourceHandler::Perl6 is SourceHandler::Raku {
-        # This may later give deprecation warnings
-    }
-    class SourceHandler::Exec does SourceHandler::Proc {
-        has @.args;
-        method new (*@args) {
-            self.bless(:@args);
-        }
-        method can-handle(Str $name) {
-            1;
-        }
-        method make-source(Str:D $name, Any:D :$err, IO::Path:D :$cwd, *%) {
-            my ($path, *@args) = @!args ?? (|@!args, $name) !! ~$cwd.add($name);
-            TAP::Source::Proc.new(:$name, :$path, :@args, :$err, :$cwd);
-        }
+    # Backwards compatibility
+    class SourceHandler {
+        constant Raku = TAP::SourceHandler::Raku;
+        constant Perl6 = TAP::SourceHandler::Raku; # will be removed later
+        constant Exec = TAP::SourceHandler::Exec;
     }
 
     my sub load-reporter(Str $name --> Reporter) {
@@ -942,7 +977,7 @@ class Harness {
 
     subset OutVal where any(IO::Handle:D, Supplier:D);
 
-    has SourceHandler @.handlers = SourceHandler::Raku.new();
+    has SourceHandlers() $.handlers = SourceHandlers.new;
     has IO::Handle $.handle = $*OUT;
     has OutVal $.output = $!handle;
     has Formatter::Volume $.volume = ?%*ENV<HARNESS_VERBOSE> ?? Verbose !! Normal;
@@ -958,14 +993,8 @@ class Harness {
     has Bool $.color;
 
     class Run does Awaitable {
-        has Promise $!waiter handles <result> is built;
-        has Promise $!bailout is built;
-        method kill(Any:D $reason) {
-            $!bailout.break($reason);
-        }
-        method get-await-handle(--> Awaitable::Handle:D) {
-            return $!waiter.promise.get-await-handle;
-        }
+        has Promise $!promise handles <result get-await-handle> is built;
+        has Promise $!bailout is built handles :kill<break>;
     }
 
     method make-aggregator() {
@@ -973,7 +1002,7 @@ class Harness {
     }
     method add-handlers(Supply $events, Output $output) {
         if $.volume == Verbose {
-            $events.act({ $output.say(~$^entry) }, :done({ $output.flush }), :quit({ $output.flush }));
+            $output.listen($events);
         }
     }
     my &sigint = sub { signal(SIGINT) }
@@ -1022,18 +1051,18 @@ class Harness {
         my $reporter = $reporter-class.new(:$output, :$formatter);
 
         my @working;
-        my $waiter = start {
+        my $promise = start {
             my $int = $!trap ?? sigint().tap({ $bailout.break("Interrupted"); $int.close(); }) !! Tap;
             my $begin = now;
             try {
                 for @names -> $name {
                     my $path = normalize-path($name, $cwd);
                     my $session = $reporter.open-test($path);
-                    my $source = @!handlers.max(*.can-handle($name)).make-source($path, :$err, :$cwd, |%handler-args);
-                    my $parser = TAP::Async.new(:$source, :$bailout, :$!loose);
+                    my $source = $!handlers.make-source($path, :$err, :$cwd, |%handler-args);
+                    my $parser = $source.parse(:$bailout, :$!loose);
                     $session.listen($parser.events);
                     self.add-handlers($parser.events, $output);
-                    @working.push({ :$parser, :$session, :done($parser.waiter) });
+                    @working.push({ :$parser, :$session, :done($parser.promise) });
                     next if @working < $!jobs;
                     await Promise.anyof(@working»<done>, $bailout);
                     reap-finished();
@@ -1065,7 +1094,7 @@ class Harness {
             }
             @working = @new-working;
         }
-        Run.new(:$waiter, :$bailout);
+        Run.new(:$promise, :$bailout);
     }
 }
 
